package edu.georgetown.gucs.dictionary;

import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.IgnoreCaseComparator;
import edu.georgetown.gucs.utility.Token;

/**
 * Thread for processing a file to add to a {@linkplain Dictionary}
 * 
 * @author Clay Shields
 * 
 **/
class DictionaryWorker implements Runnable {

  private File file;
  private Dictionary dictionary;
  private TokenizerList tokenizers;

  /**
   * Constructor that specifies a file to tokenize and add to an existing dictionary
   * 
   * @param file
   *          the file to tokenize and add to this dictionary
   * @param tokenizers
   *          the list containing the names of the tokenizers to use on this file
   * @param dictionary
   *          the dictionary to add the file to
   */
  DictionaryWorker(File file, TokenizerList tokenizers, Dictionary dictionary) {
    this.file = file;
    this.tokenizers = tokenizers;
    this.dictionary = dictionary;
  }

  /**
   * Adds a file to an existing dictionary using the given tokenizers
   */
  @Override
  public void run() {
    Global.monitor.writeToBuffer("DictionaryWorker", "run_start", "file: " + this.file + " thread_id = "
        + Thread.currentThread().getName());

    // iterate over all terms and insert them into a temporary tree map
    Map<Token, DictionaryEntry> newDictionary = Collections.synchronizedMap(new TreeMap<Token, DictionaryEntry>(
        new IgnoreCaseComparator()));

    synchronized (this.dictionary) {
      synchronized (newDictionary) {

        for (Entry<String, List<Token>> splitterEntry : this.tokenizers.tokenizeFile(this.file.getAbsolutePath())
            .entrySet()) {
          for (Token token : splitterEntry.getValue()) {
            if (token.name().length() > 0) {
              if (newDictionary.containsKey(token)) {
                newDictionary.get(token).incrementFrequencyCount();
              } else {
                DictionaryEntry dictEntry = new DictionaryEntry(token);
                newDictionary.put(token, dictEntry);
              }
            }
          }
          if (newDictionary.size() > 0) {
            this.dictionary.mergeDictionary(newDictionary);
          } else {
            Global.monitor.printErrorMessage("DictionaryWorker", "No tokens added to dictionary from " + this.file);
            Global.monitor.writeToBuffer("DictionaryWorker", "run", "No tokens added to dictionary from " + this.file);
          }
        }
      }
    }
    Global.monitor.writeToBuffer("DictionaryWorker", "run_end", "file: " + this.file + " thread_id = "
        + Thread.currentThread().getName());
  }
}