package edu.georgetown.gucs.experiment;

import java.io.File;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import edu.georgetown.gucs.matcher.ExactFingerprintMatcher;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;
import edu.georgetown.gucs.utility.Token;

/**
 * Uses a given set of parameters to create a dictionary and fingerprints of a set of documents. Matches are based on
 * comparisons of unmangled and mangled versions of each file and optionally stores the matches in a database.
 * 
 * @author Lindsay Neubauer
 * @author Clay Shields
 * 
 */
class Trial implements Runnable {

  private int trialID = -1;
  private TrialParameters base_parameters;
  private boolean finished = false;

  private Set<TrialParameters> unmangledFingerprints = Collections.synchronizedSet(new HashSet<TrialParameters>());
  private Set<TrialParameters> mangledFingerprints = Collections.synchronizedSet(new HashSet<TrialParameters>());
  private Map<String, Double[]> manglerResults = new LinkedHashMap<String, Double[]>();

  /**
   * Constructor that sets default values for this trial
   * 
   * @param trialParameters
   *          the trialParameters containing the default values for this trial
   */
  Trial(TrialParameters trialParameters) {
    this.base_parameters = trialParameters;
  }

  /**
   * Provides the parameters for this trial that are needed for results comparisons
   * 
   * @return the trialParameters containing the relevant parameters
   */
  public TrialParameters getTrialParameters() {
    return this.base_parameters;
  }

  /**
   * Provides the results for each mangler; sleeps until the trial is finished
   * 
   * @return the map containing the manglers results from this trial
   */
  public Map<String, Double[]> getManglerResults() {
    while (!this.finished) {
      try {
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    return this.manglerResults;
  }

  /**
   * Runs the trial
   */
  @Override
  public void run() {

    if (this.base_parameters.isInsert()) {
      this.trialID = this.base_parameters.getDatabase().newTrial(this.base_parameters.getRandomSeed());
      this.base_parameters.getDatabase().addTrialToExperiment(this.base_parameters.getExperimentID(), this.trialID);
    }

    if (this.base_parameters.isInsert()) {
      int dictionaryID = this.base_parameters.getDatabase().newDictionary(this.base_parameters.getDictionary());
      this.base_parameters.getDatabase().addDictionaryToTrial(this.trialID, dictionaryID);
    }

    Global.monitor.printMessage("Trial", "Setting up fingerprints");
    this.base_parameters.getFingerprinter().setDictionary(this.base_parameters.getDictionary());
    this.base_parameters.getFingerprinter().setManglerRNG(this.base_parameters.getRandom());
    this.base_parameters.getFingerprinter().setMangler(this.base_parameters.getMangler(),
        this.base_parameters.getDictionary());
    this.base_parameters.getFingerprinter().setSplitter(this.base_parameters.getSplitter());

    Global.monitor.printMessage("Trial", "Computing null fingerprints for "
        + this.base_parameters.getSampleFiles().size() + " files");
    computeFingerprints(null, null);

    Global.monitor.printMessage("Trial", "Computing splitter and mangler fingerprints:  splitter: "
        + this.base_parameters.getSplitter() + "  mangler(s)=" + this.base_parameters.getMangler());
    computeFingerprints(this.base_parameters.getMangler(), this.base_parameters.getSplitter());

    Global.monitor.printMessage("Trial", "doComparisons");
    doComparisons();
    Global.monitor.printMessage("Trial", "calculateTrialResults");
    calculateTrialResults();
  }

  /**
   * Computes fingerprints for each of the given files
   * 
   * @param manglerSettings
   *          the string containing the mangler settings to use; null if taking unmangled fingerprints
   * @param splitterString
   *          the string containing the splitter to use; null if not splitting files
   */
  private void computeFingerprints(String manglerSettings, String splitterString) {
    String mangler = manglerSettings;
    boolean unmangled = false;
    if (manglerSettings == null) {
      unmangled = true;
      mangler = "null";
    }

    Global.monitor.printMessage("Trial", "Computing fingerprints for mangler " + manglerSettings);
    Global.monitor.printMessage("Trial", "Computing fingerprints for splitter " + splitterString);
    int fileCounter = 0;
    for (File file : Collections.synchronizedList(this.base_parameters.getSampleFiles())) {
      fileCounter++;
      for (Pair<byte[], Token> fingerprintPair : this.base_parameters.getFingerprinter().computeFingerprint(
          file.getAbsolutePath())) {

        if (this.base_parameters.isInsert()) {
          int currentManglerID = this.base_parameters.getDatabase().getManglerID(mangler);
          this.base_parameters.getDatabase().addManglerToTrial(this.trialID, currentManglerID);
          this.base_parameters.getDatabase().insertFingerprint(file, fingerprintPair.getA(), this.trialID,
              currentManglerID, splitterString);
        }
        
        TrialParameters trialParameterObject = new TrialParameters(this.base_parameters);
        trialParameterObject.set(fingerprintPair.getB(), fingerprintPair.getA());

        if (unmangled) {
          this.unmangledFingerprints.add(trialParameterObject);
        } else {
          this.mangledFingerprints.add(trialParameterObject);
        }
      }
    }
    Global.monitor.writeToBuffer("Trial", "computeFingerprints", fileCounter + " files");
    Global.monitor.printMessage("Trial", "computeFingerprints: " + fileCounter + " files");
  }

  /**
   * Compares each mangled and unmangled fingerprint using this trials matcher and an exactFingerprintMatcher and
   * records true-positive, false-positive, true-negative, and false-negative results based on this comparison
   */
  private void doComparisons() {
    Global.monitor.writeToBuffer("Trial", "doComparisons", "mangled fingerprints: " + this.mangledFingerprints.size()
        + "  unmangled fingerprint: " + this.unmangledFingerprints.size());
    Global.monitor.printMessage("Trial", "Comparing " + this.mangledFingerprints.size() + " mangled fingerprints to "
        + this.unmangledFingerprints.size() + " unmangled fingerprints");

    ExactFingerprintMatcher exactMatcher = new ExactFingerprintMatcher();
    for (TrialParameters mangledParameters : this.mangledFingerprints) {
      byte[] mangledFP = mangledParameters.getFingerprint();
      String mangledFile = mangledParameters.getFileToken().name();

      for (Pair<byte[], Token> withoutMangle : this.base_parameters.getFingerprinter().computeFingerprint(mangledFile)) {
        for (TrialParameters unmangledParameters : this.unmangledFingerprints) {
          byte[] unmangledFP = unmangledParameters.getFingerprint();

          // Match and compare to an ExactFingerprint match of the unmangled fingerprints from each file
          boolean matchClaim = this.base_parameters.getMatcher().match(unmangledFP, mangledFP);
          boolean exactMatch = exactMatcher.match(unmangledFP, withoutMangle.getA());

          // Set up results
          Double[] result = this.manglerResults.get(mangledParameters.getMangler());
          if (result == null) {
            result = new Double[] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
          }

          // Update results
          if (matchClaim == true && exactMatch == true) {
            result[Global.TRUE_POSITIVE]++;
          } else if (matchClaim == true && exactMatch == false) {
            result[Global.FALSE_POSITIVE]++;
          } else if (matchClaim == false && exactMatch == false) {
            result[Global.TRUE_NEGATIVE]++;
          } else if (matchClaim == false && exactMatch == true) {
            result[Global.FALSE_NEGATIVE]++;
          }
          result[Global.DICTIONARY_SIZE] = (double) this.base_parameters.getDictionary().size();
          this.manglerResults.put(mangledParameters.getMangler(), result);

          if (this.base_parameters.isInsert()) {
            int unmangledID = Integer.valueOf(unmangledParameters.toString().substring(
                unmangledParameters.toString().charAt('@') + 1));
            int mangledID = Integer.valueOf(mangledParameters.toString().substring(
                mangledParameters.toString().charAt('@') + 1));
            int score = this.base_parameters.getMatcher().getScore(unmangledFP, mangledFP);
            this.base_parameters.getDatabase().addToFingerprintMatch(this.trialID, unmangledID, mangledID, matchClaim,
                exactMatch, score);
          }
        }
      }
    }
  }

  /**
   * Computes the precision, recall, and f-score for each mangler
   */
  private void calculateTrialResults() {
    Global.monitor.writeToBuffer("Trial", "calculateTrialResults", "");
    for (TrialParameters mangledParameters : this.mangledFingerprints) {
      Double[] result = this.manglerResults.get(mangledParameters.getMangler());

      // precision == ppv
      if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]) > 0.0) {
        result[Global.PRECISION] = result[Global.TRUE_POSITIVE]
            / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]);
      } else {
        result[Global.PRECISION] = 0.0;
      }

      // recall == sensitivity
      if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]) > 0.0) {
        result[Global.RECALL] = result[Global.TRUE_POSITIVE]
            / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]);
      } else {
        result[Global.RECALL] = 0.0;
      }

      // harmonic mean of precision and recall
      if ((result[Global.PRECISION] + result[Global.RECALL]) > 0) {
        result[Global.FSCORE] = 2 * ((result[Global.PRECISION] * result[Global.RECALL]) / (result[Global.PRECISION] + result[Global.RECALL]));
      } else {
        result[Global.FSCORE] = 0.0;
      }

      if (this.base_parameters.isInsert()) {
        Global.monitor.writeToBuffer("Trial", "calculateTrialResults", "insert into database");
        int mangledID = Integer.valueOf(mangledParameters.toString().substring(
            mangledParameters.toString().charAt('@') + 1));

        this.base_parameters.getDatabase().addTrialResults(this.trialID, mangledID,
            result[Global.TRUE_POSITIVE].intValue(), result[Global.TRUE_NEGATIVE].intValue(),
            result[Global.FALSE_POSITIVE].intValue(), result[Global.FALSE_NEGATIVE].intValue(),
            result[Global.PRECISION], result[Global.RECALL], result[Global.FSCORE]);
      }
      this.manglerResults.put(mangledParameters.getMangler(), result);
    }

    this.finished = true;
    if (this.base_parameters.isInsert()) {
      Global.monitor.printMessage("Trial", "Trial " + this.trialID + " Finished");
    } else {
      Global.monitor.printMessage("Trial", "Trial Finished");
    }
  }
}