package edu.georgetown.gucs.tokenizers;

import java.util.Collections;
import java.util.List;
import java.util.Vector;

import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Token;

/**
 * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
 * whitespace
 * 
 * @author Lindsay Neubauer
 * 
 */
public class StripMarkupTokenizer extends Tokenizer {

  private static final long serialVersionUID = -1366304768764987607L;
  private boolean keepScript = false;

  /**
   * Constructor that specifies whether to keep tokens nested inside <code><\script></code> tags; the default is to
   * eliminate these tokens
   * 
   * @param keepScript
   *          the string value (true or false) specifying whether to keep tokens nested inside <code><\script></code>
   *          tags
   */
  public StripMarkupTokenizer(String keepScript) {
    this.keepScript = Boolean.valueOf(keepScript);
  }

  /**
   * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
   * whitespace
   */
  @Override
  public List<Token> tokenize(List<Token> tokens) {
    List<Token> tokenVector = Collections.synchronizedList(new Vector<Token>());
    synchronized (tokenVector) {
      boolean script = false;
      for (Token token : tokens) {
        for (String name : token.name().split(">")) {

          // Printing error message instead of throwing exception because govdocs has unbalanced brackets
          if (name.indexOf("<") != name.lastIndexOf("<")) {
            Global.monitor.printErrorMessage("StripMarkupTokenizer", "Unbalanced brackets error: " + name + ">");
            Global.monitor
                .writeToBuffer("StripMarkupTokenizer", "tokenize", "Unbalanced brackets error: " + name + ">");
          }
          if (name.contains("script") && name.contains("<")) {
            script = !script;
          }

          if (!script) {
            if (name.indexOf('<') > 0) {
              tokenVector.add(new Token(name.substring(0, name.indexOf('<')), token.start_byte(), token.end_byte()));
            } else if (name.indexOf('<') < 0) {
              tokenVector.add(new Token(name, token.start_byte(), token.end_byte()));
            }
          } else if (script && this.keepScript) {
            tokenVector.add(new Token(name, token.start_byte(), token.end_byte()));
          }
        }
      }
      return tokenVector;
    }
  }
}
